/** @file
 * X11 visible window displayer.  Uses the same algorithm as VirtualBox's
 * seamless mode to find windows and their shapes.
 */

/*
 * Copyright (C) 2008 Oracle Corporation
 *
 * This file is part of VirtualBox Open Source Edition (OSE), as
 * available from http://www.virtualbox.org. This file is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software
 * Foundation, in version 2 as it comes in the "COPYING" file of the
 * VirtualBox OSE distribution. VirtualBox OSE is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include <set>

#include <X11/Xatom.h>
#include <X11/Xmu/WinUtil.h>
#include <X11/extensions/shape.h>

/**
 * Checks whether a window is a virtual root.
 * @returns true if it is, false otherwise
 * @param hWin the window to be examined
 */
bool isVirtualRoot(Display *pDisplay, Window hWin)
{
    unsigned char *pcWinType;
    Atom *windowType;
    bool rc = false;
    Atom actTypeAtom = None;
    int actFmt = 0;
    unsigned long nItems = 0, nBytesAfter = 0;
    unsigned char *propVal = 0;

    Atom atomNWWT = XInternAtom(pDisplay, "_NET_WM_WINDOW_TYPE", True);
    XGetWindowProperty(pDisplay, hWin, atomNWWT, 0, LONG_MAX, False, XA_ATOM,
                       &actTypeAtom, &actFmt, &nItems, &nBytesAfter, &pcWinType);
    windowType = (Atom *) pcWinType;
    if (   windowType
        && (   *windowType
            == XInternAtom(pDisplay, "_NET_WM_WINDOW_TYPE_DESKTOP", True)))
        rc = true;
    XFree(pcWinType);
    return rc;
}


int main()
{
    Display *pDisplay = XOpenDisplay(NULL);
    /** Dummy values for XShapeQueryExtension */
    int error, event;
    Bool fSupportsShape = XShapeQueryExtension(pDisplay, &event, &error);
    std::set<Window> windowsFound;
    if (!pDisplay)
    {
        printf("Unable to access the X11 window system.  Exiting.\n");
        exit(1);
    }
    Window hRoot, hParent, *phChildren;
    unsigned cChildren;
    if (!XQueryTree(pDisplay, DefaultRootWindow(pDisplay), &hRoot, &hParent,
                    &phChildren, &cChildren))
    {
        printf("Unable to enumerate active windows.  Exiting.\n");
        exit(1);
    }
    for (unsigned i = 0; i < cChildren; ++i)
    {
        Window hWin = phChildren[i];
        Window hClient = XmuClientWindow(pDisplay, hWin);
        if (!windowsFound.insert(hClient).second)
            continue;
        XWindowAttributes winAttrib;
        if (!XGetWindowAttributes(pDisplay, hWin, &winAttrib))
            continue;
        if (winAttrib.map_state == IsUnmapped)
            continue;
        XSizeHints dummyHints;
        long dummyLong;
        if (!XGetWMNormalHints(pDisplay, hClient, &dummyHints, &dummyLong))
            continue;
        if (isVirtualRoot(pDisplay, hClient))
            continue;
        XRectangle *pRects = NULL;
        int cRects = 0, iOrdering;
        bool hasShape = false;
        if (fSupportsShape)
        {
            pRects = XShapeGetRectangles(pDisplay, hWin, ShapeBounding,
                                         &cRects, &iOrdering);
            if (!pRects)
                cRects = 0;
        }
        char *pszName = NULL;
        XFetchName(pDisplay, hClient, &pszName);
        if (pszName)
            printf("Window %lu (\"%s\"): ", (unsigned long) hWin, pszName);
        else
            printf("Window %lu: ", (unsigned long) hWin);
        printf("position (%d, %d), size (%d, %d)\n", winAttrib.x, winAttrib.y,
               winAttrib.width, winAttrib.height);
        for (int j = 0; j < cRects; ++j)
            printf("    visible rectangle: position (%d, %d), size (%d, %d)\n",
                   pRects[j].x, pRects[j].y, pRects[j].width,
                   pRects[j].height);
        XFree(pRects);
        XFree(pszName);
    }
}
